/*
    Origin: ftp://ftp.jgaa.com/pub/unsupported/groupstats*.tgz
    Written by: Jarle Aase <jgaa@jgaa.com> (http://www.jgaa.com)


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 */

#define VERSION "1.00"

#include <stdio.h>
#include <unistd.h>
#include <time.h>

#define _GNU_SOURCE
#include <getopt.h>

#include <string>
#include <map>
#include <set>
#include <iostream>
#include <fstream>

#include "groupstats.h"

// Global variables
int debug_flag = 0, verbose_flag = 0;
std::string html_out = "/tmp/groupstats.html";
stats_map_t feed_map;
stats_map_t group_map;
int max_groups_in_list = 100;
int output_interval = 60; // refresh interval for reports

// Prototypes
int process(void);
void add_posting(const char *feed, int size, const char *groups, const char *distr);
void do_summary(void);
void write_stats(void);
const char *human(long long value);

void usage(void)
{
  std::cout << "Usage: groupstats [options] < feed-data" << std::endl;
  std::cout << "  Generates groupstatistics for newsfeeds" << std::endl;
  std::cout << "Options:" << std::endl;
  std::cout << "-h  --help                show this help" << std::endl;
  std::cout << "-v  --version             Print the version and exit" << std::endl;
  std::cout << "    --verbose             Verbose mode" << std::endl;
  std::cout << "    --debug               Show debug info" << std::endl;
  std::cout << "-H  --out-html=filepath   Write HTML output here" << std::endl;
  std::cout << "-m  --max-groups=#        Limit grouplist to # groups" << std::endl;
  std::cout << "-i  --out-interval=#      Minutes between reports" << std::endl;

  exit(-1);
}


int main(int argc, char *argv[])
{
  // Parse arguments
  static const struct option options[] =
  {
    {"help",    no_argument,  NULL, 'h'},
    {"verbose", no_argument,  &verbose_flag, 1},
    {"debug",   no_argument,  &debug_flag, 1},
    {"out-html", required_argument, NULL, 'H'},
    {"max-groups", required_argument, NULL, 'm'},
    {"out-interval", required_argument, NULL, 'i'},
    {"version", no_argument,  NULL, 'v'},
    {NULL,      0,            0,    0}
  };
  int c = 0;
  
  while((c = getopt_long(argc, argv, "hH:m:i:", options, NULL)) != -1)
  {
    switch(c)
    {
    case 'v':
      std::cout << "Version: " << VERSION << std::endl;
      return -1;
    case 'h':
      usage();
      break;
    case 'H':
      html_out = optarg;
      break;
    case 'm':
      max_groups_in_list = atoi(optarg);
      break;
    case 'i':
      output_interval = atoi(optarg) * 60;
      break;
    }
  }

  if (debug_flag)
    verbose_flag = 1;

  return process();
  
}

int process(void)
{
  char inbuf[1024 * 16];

  if (verbose_flag)
    std::cout << "Ready to process inncoming statistics!" << std::endl;

  // Get inndata
  while(fgets(inbuf, sizeof(inbuf), stdin))
  {
    do_summary();

    char *p = inbuf, *p_start = NULL;
    const char *feed_name = NULL, 
      *size_name = NULL, 
      *groups_name = NULL, 
      *distr_name = NULL;

    // Parse feed, size, groups, distribution
    for(p_start = p; *p && !isspace(*p); p++);
    if (*p)
    {
      feed_name = p_start;
      *p++ = 0;

      for(p_start = p; *p && !isspace(*p); p++);
      if (*p)
      {
	size_name = p_start;
	*p++ = 0;
	  
	for(p_start = p; *p && !isspace(*p); p++);
	if (*p)
	{
	  groups_name = p_start;
	  *p++ = 0;
	    
	  for(p_start = p; *p && !isspace(*p); p++);
	  {
	    if (*p_start)
	    {
	      distr_name = p_start;
	      *p = 0;
	      // Count the record
	      add_posting(feed_name, atoi(size_name), 
			  groups_name, distr_name); 
	    }  
	  }
	}
      }
    } 
  }

  write_stats();

  if (verbose_flag)
    std::cout << feed_map.size() << " feeds and " 
	      << group_map.size() << " grops seen." << std::endl;

  return 0;
}

void add_stats(stats_map_t& stats, const std::string& name, int size)
{
  stats_map_t::iterator P = stats.find(name);
  if (stats.end() == P)
    P = stats.insert(stats_t(name, Stats())).first;

  P->second.mNumPosts++;
  P->second.mNumBytes += size;
}

void add_posting(const char *feed, int size, const char *groups, const char *distr)
{
  // Count the feed
  std::string feed_name = feed, group_name;
  add_stats(feed_map, feed_name, size);

  // Count the group
  const char *p_start, *p = groups;

  while(*p)
  {
    for(p_start = p; *p && (',' != *p); p++);
    
    group_name = "";
    group_name.insert(0, p_start, p - p_start);

    if (debug_flag)
      std::clog << "Adding message to \"" << group_name << "\" " << size << " bytes" << std::endl;

    add_stats(group_map, group_name, size);

    if (*p)
      ++p;
  }
}

void do_summary(void)
{
  static time_t last_call, last_stat;
  static struct tm last_tm;

  time_t now = time(NULL);

  // Fast check to see if tha time has changed at all.
  if (last_call == now)
    return;

  // See if the day has changed...
  struct tm my_tm = *localtime(&now);

  if (0 != last_call)
  {
    //if (my_tm.tm_hour != last_tm.tm_hour)
    //  write_stats();

    if ((now - last_stat) >= output_interval)
    {
      write_stats();
      last_stat = now;
    }

    if (my_tm.tm_yday != last_tm.tm_yday)
    {
      // Reset stats...
      feed_map.erase(feed_map.begin(), feed_map.end());
      group_map.erase(group_map.begin(), group_map.end());
    }
  }
  
  last_call = now;
  last_tm = my_tm;
}

void write_stats(void)
{
  std::ofstream o;
  time_t now = time(NULL);

  if (verbose_flag)
    std::clog << "groupstats: Making HTML statistics page \"" << html_out << "\"." << std::endl;

  o.open(html_out.c_str());
  if (!o)
  {
    std::cerr << "Failed to open \"" << html_out << "\" for write!" << std::endl;
    return;
  }

  o << "<HTML><TITLE>Feed statistics</TITLE><BODY>" << std::endl;
  o << "<h1>News feed statistics page</H1>" << std::endl;
  o << "<p>Page generated " << ctime(&now) << " by Jgaa's groupstats program.</p><hr>" << std::endl;

  {
    // Sort the feeds based on volume
    stats_map_t::const_iterator PS;
    sort_set_t sorted_list;
    sort_set_t::const_iterator P;
  
    for(PS = feed_map.begin(); PS != feed_map.end(); ++PS)
      sorted_list.insert(SortedStats(PS->first.c_str(), &(PS->second)));
    
    o << "<H2>Feeds sorted by volume</H2>" << std::endl;
    o << "<TABLE border=1><TR><TD>Peer</TD><TD>Posts</TD><TD>Volume</TD></TR>" << std::endl;

    for(P = sorted_list.begin(); P != sorted_list.end(); ++P)
    {
      o << "<TR><TD>" << P->mpName << "</TD><TD align=right>" 
	<< P->mpStats->mNumPosts << "</TD><TD align=right>" 
	<< human(P->mpStats->mNumBytes) << "</TD></TR>"
	<< std::endl; 
    }

    o << "</TABLE>";
  }

 {
    // Sort the groups based on volume
    stats_map_t::const_iterator PS;
    sort_set_t sorted_list;
    sort_set_t::const_iterator P;
  
    for(PS = group_map.begin(); PS != group_map.end(); ++PS)
      sorted_list.insert(SortedStats(PS->first.c_str(), &(PS->second)));
    
    o << "<H2>Top " << max_groups_in_list << "  Groups sorted by volume</H2>" << std::endl;
    o << "<TABLE border=1><TR><TD>Group</TD><TD>Posts</TD><TD>Volume</TD></TR>" << std::endl;

    int count = 0;
    for(P = sorted_list.begin(); P != sorted_list.end(); ++P)
    {
      if (max_groups_in_list && (count++ > max_groups_in_list))
	break;

      o << "<TR><TD>" << P->mpName << "</TD><TD align=right>" 
	<< P->mpStats->mNumPosts << "</TD><TD align=right>" 
	<< human(P->mpStats->mNumBytes) << "</TD></TR>"
	<< std::endl; 
    }

    o << "</TABLE>";
  }

 {
    // Sort the groups based on number of posts
    stats_map_t::const_iterator PS;
    sort_set_t sorted_list;
    sort_set_t::const_iterator P;
  
    for(PS = group_map.begin(); PS != group_map.end(); ++PS)
      sorted_list.insert(SortedStats(PS->first.c_str(), &(PS->second), true));
    
    o << "<H2>Top " << max_groups_in_list << "  Groups sorted by posts</H2>" << std::endl;
    o << "<TABLE border=1><TR><TD>Group</TD><TD>Posts</TD><TD>Volume</TD></TR>" << std::endl;

    int count = 0;
    for(P = sorted_list.begin(); P != sorted_list.end(); ++P)
    {
      if (max_groups_in_list && (count++ > max_groups_in_list))
	break;

      o << "<TR><TD>" << P->mpName << "</TD><TD align=right>" 
	<< P->mpStats->mNumPosts << "</TD><TD align=right>" 
	<< human(P->mpStats->mNumBytes) << "</TD></TR>"
	<< std::endl; 
    }

    o << "</TABLE>";
  }

 
  o << "</BODY></HTML>" << std::endl;

}

const char *human(long long value)
{
  static char buf[512];

  // Terrabytes
  if (value > (1024LL * 1024 * 1024 * 1024))
  {
    sprintf(buf, "%0.2f Tb", (value / (1024.0 * 1024.0 *1024.0 * 1024.0)));
    return buf;
  }

  // Gigabytes
  if (value > (1024LL * 1024 * 1024))
  {
    sprintf(buf, "%0.2f Gb", (value / (1024.0 * 1024.0 *1024.0)));
    return buf;
  }

  // Megabytes
  if (value > (1024 * 1024))
  {
    sprintf(buf, "%0.2f Mb", (value / (1024.0 * 1024.0)));
    return buf;
  }

  // Kilobytes
  if (value > (1024))
  {
    sprintf(buf, "%0.2f Kb", (value / 1024.0));
    return buf;
  }

  sprintf (buf, "%d bytes", (int)value);
  return buf;
}
